#!/usr/bin/env python
"""
_WMWorkload_

Request level processing specification, acts as a container of a set
of related tasks.
"""
from __future__ import print_function

from builtins import next, range
from future.utils import viewitems, viewvalues
from collections  import namedtuple
import inspect


from Utils.Utilities import strToBool, makeList
from WMCore.Configuration import ConfigSection
from WMCore.Lexicon import sanitizeURL
from WMCore.WMException import WMException
from WMCore.WMSpec.WMSpecErrors import WMSpecFactoryException
from WMCore.WMSpec.ConfigSectionTree import findTop
from WMCore.WMSpec.Persistency import PersistencyHelper
from WMCore.WMSpec.WMTask import WMTask, WMTaskHelper
from WMCore.WMSpec.WMWorkloadTools import (validateArgumentsUpdate, validateUnknownArgs, validateSiteLists,
                                           _validateArgumentOptions, loadSpecClassByType,
                                           setAssignArgumentsWithDefault)

parseTaskPath = lambda p: [x for x in p.split('/') if x.strip() != '']


def getWorkloadFromTask(taskRef):
    """
    _getWorkloadFromTask_

    Util to retrieve a Workload wrapped in a WorkloadHelper
    from a WMTask.
    """
    nodeData = taskRef
    if isinstance(taskRef, WMTaskHelper):
        nodeData = taskRef.data

    topNode = findTop(nodeData)
    if not hasattr(topNode, "objectType"):
        msg = "Top Node is not a WM definition object:\n"
        msg += "Object has no objectType attribute"
        # TODO: Replace with real exception class
        raise RuntimeError(msg)

    objType = getattr(topNode, "objectType")
    if objType != "WMWorkload":
        msg = "Top level object is not a WMWorkload: %s" % objType
        # TODO: Replace with real exception class
        raise RuntimeError(msg)

    return WMWorkloadHelper(topNode)


class WMWorkloadException(WMException):
    """
    _WMWorkloadException_

    Exceptions raised by the Workload during filling
    """
    pass

class WMWorkloadUnhandledException(WMException):
    """
    _WMWorkloadUnhandledException_

    Exceptions raised by the Workload during filling
    """
    pass


setterTuple = namedtuple('SetterTuple', ['reqArg', 'setterFunc', 'setterSignature'])

class WMWorkloadHelper(PersistencyHelper):
    """
    _WMWorkloadHelper_

    Methods & Utils for working with a WMWorkload instance.
    """

    def __init__(self, wmWorkload=None):
        self.data = wmWorkload
        self.settersMap = {}

    def updateWorkloadArgs(self, reqArgs):
        """
        Method to take a dictionary of arguments of the type:
        {reqArg1: value,
         reqArg2: value,
         ...}
        and update the workload by a predefined map of reqArg to setter methods.
        :param reqArgs: A Dictionary of request arguments to be updated
        :return:        Nothing, Raises an error of type WMWorkloadException if
                        fails to apply the proper setter method
        """
        # NOTE: So far we support only a single argument setter methods, like
        #       setSiteWhitelist or setPriority. This may change in the future,
        #       but it will require a change in the logic of how we validate and
        #       call the proper setter methods bellow.

        # populate the current instance settersMap
        self.settersMap['RequestPriority'] = setterTuple('RequestPriority', self.setPriority, inspect.signature(self.setPriority))
        self.settersMap['SiteBlacklist'] = setterTuple('SiteBlacklist', self.setSiteBlacklist, inspect.signature(self.setSiteBlacklist))
        self.settersMap['SiteWhitelist'] = setterTuple('SiteWhitelist', self.setSiteWhitelist, inspect.signature(self.setSiteWhitelist))

        reqArgsNothandled = []
        # First validate if we can properly call the setter function given the reqArgs passed.
        for reqArg, argValue in reqArgs.items():
            if reqArg not in self.settersMap:
                reqArgsNothandled.append(reqArg)
                continue
            try:
                self.settersMap[reqArg].setterSignature.bind(argValue)
            except TypeError as ex:
                msg = f"Setter's method signature does not match the method calls we currently support: Error: req{str(ex)}"
                raise WMWorkloadException(msg) from None

        if reqArgsNothandled:
            msg = f"Unsupported or missing setter method for updating request arguments: {reqArgsNothandled}."
            raise WMWorkloadUnhandledException(msg) from None

        # Now go through the reqArg again and call every setter method according to the map
        for reqArg, argValue in reqArgs.items():
            try:
                self.settersMap[reqArg].setterFunc(argValue)
            except Exception as ex:
                currFrame = inspect.currentframe()
                argsInfo = inspect.getargvalues(currFrame)
                argVals = {arg: argsInfo.locals.get(arg) for arg in argsInfo.args}
                msg = f"Failure while calling setter method {self.settersMap[reqArg].setterFunc.__name__} "
                msg += f"With arguments: {argVals}"
                msg += f"Full exception string: {str(ex)}"
                raise WMWorkloadException(msg) from None

    def setSpecUrl(self, url):
        self.data.persistency.specUrl = sanitizeURL(url)["url"]

    def specUrl(self):
        """
        _specUrl_

        return url location of workload
        """
        return self.data.persistency.specUrl

    def name(self):
        """
        _name_

        return name of the workload
        """
        return self.data._internal_name

    def setName(self, workloadName):
        """
        _setName_

        Set the workload name.
        """
        self.data._internal_name = workloadName
        return

    def setRequestType(self, requestType):
        self.data.requestType = requestType

    def setStepProperties(self, assignArgs):
        """
        _setStepProperties_

        Used for properly setting AcqEra/ProcStr/ProcVer for each step in a StepChain request
        during assignment. Only used if one of those parameters is a dictionary.
        """
        if "AcquisitionEra" in assignArgs and isinstance(assignArgs["AcquisitionEra"], dict):
            pass
        elif "ProcessingString" in assignArgs and isinstance(assignArgs["ProcessingString"], dict):
            pass
        elif "ProcessingVersion" in assignArgs and isinstance(assignArgs["ProcessingVersion"], dict):
            pass
        else:
            return

        stepNameMapping = self.getStepMapping()
        # it has only one top level task
        for task in self.taskIterator():
            # Merge task has cmsRun1 step, so it gets messy on Merge ACDC of StepChain
            if task.taskType() == "Merge":
                continue
            task.updateLFNsAndDatasets(dictValues=assignArgs, stepMapping=stepNameMapping)

        return

    def setTaskEnvironmentVariables(self, envDict):
        """
        _setTaskEnvironmentVariables_

        Used for setting environment variables for each task in a request.
        """
        if not isinstance(envDict, dict):
            return

        for task in self.taskIterator():
            task.addEnvironmentVariables(envDict)
        return

    def setOverrideCatalog(self, tfcFile):
        """
        _setOverrideCatalog_

        Used for setting overrideCatalog option for each step in the workload.
        """
        for task in self.taskIterator():
            task.setOverrideCatalog(tfcFile)
        return

    def setStepMapping(self, mapping):
        """
        _setStepMapping_

        Mostly used for StepChains. It creates a mapping between the StepName and the step
        number and the cmsRun number. E.g.:
          {'GENSIM': ('Step1', 'cmsRun1'), 'DIGI': ('Step2', 'cmsRun2'), 'RECO': ('Step3', 'cmsRun3')}
        """
        self.data.properties.stepMapping = mapping

    def getStepMapping(self):
        """
        _getStepMapping_

        Only important for StepChains. Map from step name to step number
        and cmsRun number.
        """
        return getattr(self.data.properties, "stepMapping", None)

    def setStepParentageMapping(self, mapping):
        """
        _setStepParentageMapping_

        Used for StepChains. Set a wider dictionary structure with a mapping between
        parent and child steps as well as dataset parentage
        """
        self.data.properties.stepParentageMapping = mapping

    def getStepParentageMapping(self):
        """
        _getStepParentageMapping_

        Only important for StepChains. Map from step name to step and parent
        step properties, including a map of output datasets to the parent dataset.
        """
        return getattr(self.data.properties, "stepParentageMapping", {})

    def getStepParentDataset(self, childDataset):
        """
        :param childDataset: child dataset which is looking for parent dataset
        :return: str parent dataset if exist, otherwise None

        Correct parentage mapping is set when workflow is assigned, Shouldn't call this method before workflow is assigned
        Assumes there is only one parent dataset given childDataset
        """
        ### FIXME: Seangchan, I don't think we need this method, since we'll add the
        # map to the dbsbuffer_dataset table and then use it from there. So,
        # wmbsHelper should actually fetch the simple map data and insert that into db
        stepParentageMap = self.getStepParentageMapping()
        if stepParentageMap:
            for stepName in stepParentageMap:
                stepItem = stepParentageMap[stepName]
                outDSMap = stepItem["OutputDatasetMap"]
                for outmodule in outDSMap:
                    if childDataset in outDSMap[outmodule] and stepItem['ParentDataset']:
                        return stepItem['ParentDataset']
        else:
            return None

    def setTaskParentageMapping(self, mapping):
        """
        _setTaskParentageMapping_

        Used for TaskChains. Sets a dictionary with the task / parent task /
        parent dataset / and output datasets relationship.
        """
        self.data.properties.taskParentageMapping = mapping

    def getTaskParentageMapping(self):
        """
        _getTaskParentageMapping_

        Only important for TaskChains. Returns a map of task name to
        parent dataset and output datasets.
        """
        return getattr(self.data.properties, "taskParentageMapping", {})

    def getChainParentageSimpleMapping(self):
        """
        Creates a simple map of task or step to parent and output datasets
        such that it can be friendly stored in the reqmgr workload cache doc.
        :return:  {'Step1': {'ParentDset': 'blah1', 'ChildDsets': ['blah2']},
                   'Step2': {'ParentDset': 'blah2', 'ChildDsets': ['blah3', 'blah4],
                   ...} if stepParentageMapping exist otherwise None
        """
        if self.getRequestType() == "TaskChain":
            chainMap = self.getTaskParentageMapping()
        elif self.getRequestType() == "StepChain":
            chainMap = self.getStepParentageMapping()
        else:
            return {}

        newMap = {}
        if chainMap:
            for cData in viewvalues(chainMap):
                cNum = cData.get('TaskNumber', cData.get('StepNumber'))
                newMap[cNum] = {'ParentDset': cData['ParentDataset'],
                                'ChildDsets': []}
                for outMod in cData['OutputDatasetMap']:
                    newMap[cNum]['ChildDsets'].append(cData['OutputDatasetMap'][outMod])
        return newMap

    def updateStepParentageMap(self):
        """
        _updateStepParentageMap
        Used to update the step parentage mapping of StepChain requests at the
        end of the assignment process, given that we might have new output
        dataset names
        :return: just updates the workload property: stepParentageMapping
        """
        topLevelTask = next(self.taskIterator())
        if topLevelTask.taskType() == "Merge":
            # handle ACDC for merge jobs, see #9051. Nothing to do here
            return

        parentMap = self.getStepParentageMapping()
        listOfStepNames = list(parentMap)
        for stepName in listOfStepNames:
            if parentMap[stepName]['OutputDatasetMap']:
                # then there is output dataset, let's update it
                cmsRunNumber = parentMap[stepName]['StepCmsRun']
                stepHelper = topLevelTask.getStepHelper(cmsRunNumber)
                for outputModuleName in stepHelper.listOutputModules():
                    outputModule = stepHelper.getOutputModule(outputModuleName)
                    outputDataset = "/%s/%s/%s" % (outputModule.primaryDataset,
                                                   outputModule.processedDataset,
                                                   outputModule.dataTier)

                    # now find and replace the old dataset by the new dataset name
                    oldOutputDset = parentMap[stepName]['OutputDatasetMap'][outputModuleName]
                    for s in listOfStepNames:
                        if parentMap[s]['ParentDataset'] == oldOutputDset:
                            parentMap[s]['ParentDataset'] = outputDataset
                        if oldOutputDset == parentMap[s]['OutputDatasetMap'].get(outputModuleName, ""):
                            parentMap[s]['OutputDatasetMap'][outputModuleName] = outputDataset

        self.setStepParentageMapping(parentMap)

        return

    def updateTaskParentageMap(self):
        """
        _updateTaskParentageMap_
        Used to update the task dataset parentage mapping of TaskChain requests
        at the end of the assignment process, given that we might have new output
        dataset names
        :return: just updates the workload property: taskParentageMapping
        """
        taskMap = self.getTaskParentageMapping()

        for tName in taskMap:
            if not taskMap[tName]['OutputDatasetMap']:
                continue

            taskO = self.getTaskByName(tName)
            if taskO is None:
                # Resubmission requests might not have certain tasks
                continue

            for outInfo in taskO.listOutputDatasetsAndModules():
                # Check whether it's a transient output module
                if outInfo['outputModule'] not in taskMap[tName]['OutputDatasetMap']:
                    continue
                oldOutputDset = taskMap[tName]['OutputDatasetMap'][outInfo['outputModule']]
                taskMap[tName]['OutputDatasetMap'][outInfo['outputModule']] = outInfo['outputDataset']
                for tt in taskMap:
                    if taskMap[tt]['ParentDataset'] == oldOutputDset:
                        taskMap[tt]['ParentDataset'] = outInfo['outputDataset']

        self.setTaskParentageMapping(taskMap)

        return

    def getInitialJobCount(self):
        """
        _getInitialJobCount_

        Get the initial job count, this is incremented everytime the workflow
        is resubmitted with ACDC.
        """
        return self.data.initialJobCount

    def setInitialJobCount(self, jobCount):
        """
        _setInitialJobCount_

        Set the initial job count.
        """
        self.data.initialJobCount = jobCount
        return

    def getDashboardActivity(self):
        """
        _getDashboardActivity_

        Retrieve the dashboard activity.
        """
        return self.data.properties.dashboardActivity

    def setDashboardActivity(self, dashboardActivity):
        """
        _setDashboardActivity_

        Set the dashboard activity for the workflow.
        """
        self.data.properties.dashboardActivity = dashboardActivity
        return

    def getTopLevelTask(self):
        """
        _getTopLevelTask_

        Retrieve the top level task.
        """
        topLevelTasks = []
        for task in self.taskIterator():
            if task.isTopOfTree():
                topLevelTasks.append(task)

        return topLevelTasks

    def getOwner(self):
        """
        _getOwner_

        Retrieve the owner information.
        """
        return self.data.owner.dictionary_()

    def setOwner(self, name, ownerProperties=None):
        """
        _setOwner_
        sets the owner of wmspec.
        Takes a name as a mandatory argument, and then a dictionary of properties
        """
        ownerProperties = ownerProperties or {'dn': 'DEFAULT'}

        self.data.owner.name = name
        self.data.owner.group = "undefined"

        if not isinstance(ownerProperties, dict):
            raise Exception("Someone is trying to setOwner without a dictionary")

        for key in ownerProperties:
            setattr(self.data.owner, key, ownerProperties[key])

        return

    def setOwnerDetails(self, name, group, ownerProperties=None):
        """
        _setOwnerDetails_

        Set the owner, explicitly requiring the group and user arguments
        """
        ownerProperties = ownerProperties or {'dn': 'DEFAULT'}

        self.data.owner.name = name
        self.data.owner.group = group

        if not isinstance(ownerProperties, dict):
            raise Exception("Someone is trying to setOwnerDetails without a dictionary")
        for key in ownerProperties:
            setattr(self.data.owner, key, ownerProperties[key])
        return

    def sandbox(self):
        """
        _sandbox_
        """
        return self.data.sandbox

    def setSandbox(self, sandboxPath):
        """
        _sandbox_
        """
        self.data.sandbox = sandboxPath

    def setPriority(self, priority):
        """
        _setPriority_

        Set the priority for the workload
        """
        self.data.request.priority = int(priority)

    def priority(self):
        """
        _priority_
        return priority of workload
        """
        return self.data.request.priority

    def setStartPolicy(self, policyName, **params):
        """
        _setStartPolicy_

        Set the Start policy and its parameters
        """
        self.data.policies.start.policyName = policyName
        for key, val in viewitems(params):
            setattr(self.data.policies.start, key, val)

    def startPolicy(self):
        """
        _startPolicy_

        Get Start Policy name
        """
        return getattr(self.data.policies.start, "policyName", None)

    def startPolicyParameters(self):
        """
        _startPolicyParameters_

        Get Start Policy parameters
        """
        datadict = getattr(self.data.policies, "start")
        return datadict.dictionary_()

    def setEndPolicy(self, policyName, **params):
        """
        _setEndPolicy_

        Set the End policy and its parameters
        """
        self.data.policies.end.policyName = policyName
        for key, val in viewitems(params):
            setattr(self.data.policies.end, key, val)

    def endPolicy(self):
        """
        _endPolicy_

        Get End Policy name
        """
        return getattr(self.data.policies.end, "policyName", None)

    def endPolicyParameters(self):
        """
        _startPolicyParameters_

        Get Start Policy parameters
        """
        datadict = getattr(self.data.policies, "end")
        return datadict.dictionary_()

    def getTask(self, taskName):
        """
        _getTask_

        Retrieve a - top level task - with the given name.
        """
        task = getattr(self.data.tasks, taskName, None)
        if task is None:
            return None
        return WMTaskHelper(task)

    def getTaskByName(self, taskName):
        """
        _getTaskByName_

        Retrieve a task with the given name in the whole workflow tree.
        """
        for t in self.taskIterator():
            if t.name() == taskName:
                return t
            for x in t.taskIterator():
                if x.name() == taskName:
                    return x
        return None

    def getTaskByPath(self, taskPath):
        """
        _getTask_

        Get a task instance based on the path name

        """
        mapping = {}
        for t in self.taskIterator():
            for x in t.taskIterator():
                mapping.__setitem__(x.getPathName, x.name())

        taskList = parseTaskPath(taskPath)

        if taskList[0] != self.name():  # should always be workload name first
            msg = "Workload name does not match:\n"
            msg += "requested name %s from workload %s " % (taskList[0],
                                                            self.name())
            raise RuntimeError(msg)
        if len(taskList) < 2:
            # path should include workload and one task
            msg = "Task Path does not contain a top level task:\n"
            msg += taskPath
            raise RuntimeError(msg)

        topTask = self.getTask(taskList[1])
        if topTask is None:
            msg = "Task /%s/%s Not Found in Workload" % (taskList[0],
                                                         taskList[1])
            raise RuntimeError(msg)
        for x in topTask.taskIterator():
            if x.getPathName() == taskPath:
                return x
        return None

    def taskIterator(self):
        """
        generator to traverse top level tasks

        """
        for i in self.data.tasks.tasklist:
            yield self.getTask(i)

    def listAllTaskNodes(self):
        """
        """
        result = []
        for t in self.taskIterator():
            if t != None:
                result.extend(t.listNodes())
        return result

    def listAllTaskPathNames(self):
        """
        _listAllTaskPathNames_

        Generate a list of all known task path names including
        tasks that are part of the top level tasks
        """
        result = []
        for t in self.taskIterator():
            result.extend(t.listPathNames())
        return result

    def listAllTaskNames(self):
        """
        _listAllTaskNames_

        Generate a list of all known task names including
        tasks that are part of the top level tasks
        """
        result = []
        for t in self.taskIterator():
            result.extend(t.listNames())
        return result

    def listTasksOfType(self, ttype):
        """
        _listTasksOfType_

        Get tasks matching the type provided
        """
        return [t for t in self.taskIterator() if t.taskType() == ttype]

    def getAllTasks(self, cpuOnly=False):
        """
        _getAllTasks_

        Get all tasks from a workload.
        If cpuOnly flag is set to True, then don't return utilitarian tasks.
        """
        tasks = []
        for n in self.listAllTaskPathNames():
            task = self.getTaskByPath(taskPath=n)
            if cpuOnly and task.taskType() in ["Cleanup", "LogCollect"]:
                continue
            tasks.append(task)

        return tasks

    def addTask(self, wmTask):
        """
        _addTask_

        Add a Task instance either naked or wrapped in a helper

        """
        task = wmTask
        if isinstance(wmTask, WMTaskHelper):
            task = wmTask.data
            helper = wmTask
        else:
            helper = WMTaskHelper(wmTask)
        taskName = helper.name()
        pathName = "/%s/%s" % (self.name(), taskName)
        helper.setPathName(pathName)
        if taskName in self.listAllTaskNodes():
            msg = "Duplicate task name: %s\n" % taskName
            msg += "Known tasks: %s\n" % self.listAllTaskNodes()
            raise RuntimeError(msg)
        self.data.tasks.tasklist.append(taskName)
        setattr(self.data.tasks, taskName, task)
        return

    def newTask(self, taskName):
        """
        _newTask_

        Factory like interface for adding a toplevel task to this
        workload

        """
        if taskName in self.listAllTaskNodes():
            msg = "Duplicate task name: %s\n" % taskName
            msg += "Known tasks: %s\n" % self.listAllTaskNodes()
            raise RuntimeError(msg)
        task = WMTask(taskName)
        helper = WMTaskHelper(task)
        helper.setTopOfTree()
        self.addTask(helper)
        return helper

    def removeTask(self, taskName):
        """
        _removeTask_

        Remove given task with given name

        """
        self.data.tasks.__delattr__(taskName)
        self.data.tasks.tasklist.remove(taskName)
        return

    def getSiteWhitelist(self):
        """
        Get the site white list for the workflow
        :return: site white list
        """
        # loop over tasks to see if there is white lists
        taskIterator = self.taskIterator()
        siteList = []
        for task in taskIterator:
            for site in task.siteWhitelist():
                siteList.append(site)
        return list(set(siteList))

    def setSiteWhitelist(self, siteWhitelist):
        """
        _setSiteWhitelist_

        Set the site white list for the top level tasks in the workload.
        """
        if not isinstance(siteWhitelist, list):
            siteWhitelist = [siteWhitelist]

        taskIterator = self.taskIterator()

        for task in taskIterator:
            task.setSiteWhitelist(siteWhitelist)

        return

    def getSiteBlacklist(self):
        """
        Get the site black list for the workflow
        :return: site black list
        """
        # loop over tasks to see if there is black lists
        taskIterator = self.getAllTasks(cpuOnly=False)
        siteList = []
        for task in taskIterator:
            for site in task.siteBlacklist():
                siteList.append(site)
        return list(set(siteList))

    def setSiteBlacklist(self, siteBlacklist):
        """
        _setSiteBlacklist_

        Set the site black list for all tasks in the workload.
        """
        if not isinstance(siteBlacklist, type([])):
            siteBlacklist = [siteBlacklist]

        taskIterator = self.getAllTasks(cpuOnly=False)

        for task in taskIterator:
            task.setSiteBlacklist(siteBlacklist)

        return

    def setBlockWhitelist(self, blockWhitelist, initialTask=None):
        """
        _setBlockWhitelist_

        Set the block white list for all tasks that have an input dataset
        defined.
        """
        if not isinstance(blockWhitelist, type([])):
            blockWhitelist = [blockWhitelist]

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            if task.getInputDatasetPath():
                task.setInputBlockWhitelist(blockWhitelist)
            self.setBlockWhitelist(blockWhitelist, task)

        return

    def setBlockBlacklist(self, blockBlacklist, initialTask=None):
        """
        _setBlockBlacklist_

        Set the block black list for all tasks that have an input dataset
        defined.
        """
        if not isinstance(blockBlacklist, type([])):
            blockBlacklist = [blockBlacklist]

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            if task.getInputDatasetPath():
                task.setInputBlockBlacklist(blockBlacklist)
            self.setBlockBlacklist(blockBlacklist, task)

        return

    def setRunWhitelist(self, runWhitelist, initialTask=None):
        """
        _setRunWhitelist_

        Set the run white list for all tasks that have an input dataset defined.
        """
        if not isinstance(runWhitelist, type([])):
            runWhitelist = [runWhitelist]

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            if task.getInputDatasetPath():
                task.setInputRunWhitelist(runWhitelist)
                task.setSplittingParameters(runWhitelist=runWhitelist)
            self.setRunWhitelist(runWhitelist, task)

        return

    def setRunBlacklist(self, runBlacklist, initialTask=None):
        """
        _setRunBlacklist_

        Set the run black list for all tasks that have an input dataset defined.
        """
        if not isinstance(runBlacklist, type([])):
            runBlacklist = [runBlacklist]

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            if task.getInputDatasetPath():
                task.setInputRunBlacklist(runBlacklist)
            self.setRunBlacklist(runBlacklist, task)

        return

    def updateLFNsAndDatasets(self, runNumber=None):
        """
        _updateLFNsAndDatasets_

        Update all the output LFNs and data names for all tasks in the workflow.
        This needs to be called after updating the acquisition era, processing
        version or merged/unmerged lfn base.
        """
        taskIterator = self.taskIterator()

        for task in taskIterator:
            task.updateLFNsAndDatasets(runNumber=runNumber)
        return

    def updateDatasetName(self, mergeTask, datasetName):
        """
        _updateDatasetName_

        Updates the dataset name argument of the mergeTask's harvesting
        children tasks
        """
        for task in mergeTask.childTaskIterator():
            if task.taskType() == "Harvesting":
                for stepName in task.listAllStepNames():
                    stepHelper = task.getStepHelper(stepName)

                    if stepHelper.stepType() == "CMSSW":
                        cmsswHelper = stepHelper.getTypeHelper()
                        cmsswHelper.setDatasetName(datasetName)

        return

    def setCoresAndStreams(self, cores, nStreams, initialTask=None):
        """
        _setCoresAndStreams_

        Update the number of cores and event streams for each task in the spec.

        One can update only the number of cores, which will set the number of streams to 0 (default).
        However, updating only the number of streams is not allowed, it's coupled to # of cores.

        :param cores: number of cores. Can be either an integer or a dict key'ed by taskname
        :param nStreams: number of streams. Can be either an integer or a dict key'ed by taskname
        :param initialTask: parent task object
        """
        if not cores:
            return

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            task.setNumberOfCores(cores, nStreams)
            self.setCoresAndStreams(cores, nStreams, task)

        return

    def setGPUSettings(self, requiresGPU, gpuParams, initialTask=None):
        """
        Setter method for the workload GPU parameters.
        It's responsible for setting whether GPUs are required or not; and
        which GPU parameters to be used for that. This is done for every
        task of this spec.
        :param requiresGPU: string defining whether GPUs are needed. For TaskChains, it
            could be a dictionary key'ed by the taskname.
        :param gpuParams: GPU settings. A JSON encoded object, from either a None object
            or a dictionary. For TaskChains, it could be a dictionary key'ed by the taskname
        :param initialTask: parent task object
        """
        if not requiresGPU:
            return

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            task.setTaskGPUSettings(requiresGPU, gpuParams)
            self.setGPUSettings(requiresGPU, gpuParams, task)
        return

    def setMemory(self, memory, initialTask=None):
        """
        _setMemory_

        Update memory requirements for each task in the spec, thus it
        can be either an integer or a dictionary key'ed by the task name.
        """
        if not memory:
            return

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            if isinstance(memory, dict):
                mem = memory.get(task.name())
            else:
                mem = memory
            task.setJobResourceInformation(memoryReq=mem)
            self.setMemory(memory, task)

        return

    def setTimePerEvent(self, timePerEvent, initialTask=None):
        """
        _setTimePerEvent_

        Update TimePerEvent requirements for each task in the spec, thus it
        can be either an integer or a dictionary key'ed by the task name.
        """
        # don't set it for utilitarian/merge tasks
        if not timePerEvent:
            return

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            if isinstance(timePerEvent, dict):
                timePE = timePerEvent.get(task.name())
            else:
                timePE = timePerEvent
            task.setJobResourceInformation(timePerEvent=timePE)
            self.setTimePerEvent(timePerEvent, task)

        return

    def setAcquisitionEra(self, acquisitionEras):
        """
        _setAcquistionEra_

        Change the acquisition era for all tasks in the spec and then update
        all of the output LFNs and datasets to use the new acquisition era.
        """
        stepNameMapping = self.getStepMapping()
        for task in self.taskIterator():
            task.setAcquisitionEra(acquisitionEras, stepChainMap=stepNameMapping)

        self.updateLFNsAndDatasets()
        # set acquistionEra for workload (need to refactor)
        self.data.properties.acquisitionEra = acquisitionEras
        return

    def setProcessingVersion(self, processingVersions):
        """
        _setProcessingVersion_

        Change the processing version for all tasks in the spec and then update
        all of the output LFNs and datasets to use the new processing version.

        :param processingVersions: can be any data-type but it is set from StdBase
        which performs the input data sanitization/type already.
        """
        stepNameMapping = self.getStepMapping()

        for task in self.taskIterator():
            task.setProcessingVersion(processingVersions, stepChainMap=stepNameMapping)

        self.updateLFNsAndDatasets()
        self.data.properties.processingVersion = processingVersions
        return

    def setProcessingString(self, processingStrings):
        """
        _setProcessingString_

        Change the processing string for all tasks in the spec and then update
        all of the output LFNs and datasets to use the new processing version.
        """
        stepNameMapping = self.getStepMapping()

        for task in self.taskIterator():
            task.setProcessingString(processingStrings, stepChainMap=stepNameMapping)

        self.updateLFNsAndDatasets()
        self.data.properties.processingString = processingStrings
        return

    def setLumiList(self, lumiLists):
        """
        _setLumiList_

        Change the lumi mask for all tasks in the spec
        """

        for task in self.taskIterator():
            task.setLumiMask(lumiLists, override=False)

        # set lumiList for workload (need to refactor)
        self.data.properties.lumiList = lumiLists
        return

    def setTaskProperties(self, requestArgs):
        # FIXME (Alan): I don't think it works, given that the assignment
        # parameters never have the TaskChain parameter...
        if not 'TaskChain' in requestArgs:
            return
        numTasks = requestArgs['TaskChain']
        taskArgs = []
        for i in range(numTasks):
            taskArgs.append(requestArgs["Task%s" % (i + 1)])

        for prop in taskArgs:
            taskName = prop['TaskName']
            for task in self.getAllTasks():
                if task.name() == taskName:
                    del prop['TaskName']
                    task.setProperties(prop)
                    break
        return

    def getAcquisitionEra(self, taskName=None):
        """
        _getAcquisitionEra_

        Get the acquisition era
        """
        if taskName and isinstance(self.data.properties.acquisitionEra, dict):
            return self.data.properties.acquisitionEra.get(taskName, None)
        return self.data.properties.acquisitionEra

    def getRequestType(self):
        """
        _getRequestType_

        Get the Request type (ReReco, TaskChain, etc)
        """
        if getattr(self.data, 'requestType', None):
            return getattr(self.data, "requestType")

        if hasattr(self.data, "request"):
            if hasattr(self.data.request, "schema"):
                return getattr(self.data.request.schema, "RequestType", None)
        return None

    def getProcessingVersion(self, taskName=None):
        """
        _getProcessingVersion_

        Get the processingVersion
        """

        if taskName and isinstance(self.data.properties.processingVersion, dict):
            return self.data.properties.processingVersion.get(taskName, 0)
        return self.data.properties.processingVersion

    def getProcessingString(self, taskName=None):
        """
        _getProcessingString_

        Get the processingString
        """

        if taskName and isinstance(self.data.properties.processingString, dict):
            return self.data.properties.processingString.get(taskName, None)
        return self.data.properties.processingString

    def getLumiList(self):
        """
        _getLumiList_

        Get the LumitList from workload (task level should have the same lumiList)
        """
        return self.data.properties.lumiList

    def setValidStatus(self, validStatus):
        """
        _setValidStatus_

        Sets the status that will be reported to the processed dataset
        in DBS
        """

        self.data.properties.validStatus = validStatus
        return

    def getValidStatus(self):
        """
        _getValidStatus_

        Get the valid status for DBS
        """

        return getattr(self.data.properties, 'validStatus', None)

    def setAllowOpportunistic(self, allowOpport):
        """
        _setAllowOpportunistic_

        Set a flag which enables the workflow to run in cloud resources.
        """
        self.data.properties.allowOpportunistic = allowOpport
        return

    def getAllowOpportunistic(self):
        """
        _getAllowOpportunistic_

        Retrieve AllowOpportunitisc flag for the workflow
        """
        return getattr(self.data.properties, 'allowOpportunistic', None)

    def setPrepID(self, prepID):
        """
        _setPrepID_

        Set the prepID to for all the tasks below
        """

        taskIterator = self.taskIterator()
        for task in taskIterator:
            task.setPrepID(prepID)
        self.data.properties.prepID = prepID
        return

    def getPrepID(self):
        """
        _getPrepID_

        Get the prepID for the workflow
        """
        return getattr(self.data.properties, 'prepID', None)

    def setDbsUrl(self, dbsUrl):
        """
        _setDbsUrl_

        Set the workload level DbsUrl.
        """
        # TODO: this replace can be removed in one year from now, thus March 2022
        dbsUrl = dbsUrl.replace("cmsweb.cern.ch", "cmsweb-prod.cern.ch")
        # stripping any end slashes, which no longer work in the Go-based server
        self.data.dbsUrl = dbsUrl.rstrip("/")

    def getDbsUrl(self):
        """
        _getDbsUrl_

        Get the DbsUrl specified for the input dataset.
        """
        if getattr(self.data, 'dbsUrl', None):
            return getattr(self.data, "dbsUrl")

        if hasattr(self.data, "request"):
            if hasattr(self.data.request, "schema"):
                if not getattr(self.data.request.schema, "DbsUrl", None):
                    return "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader"

        return getattr(self.data.request.schema, "DbsUrl")

    def setStatus(self, status):
        """
        _setStatus_

        Set the status of the workflow
        Optional
        :param: Request status
        """
        self.data.request.status = status
        return

    def getStatus(self):
        """
        _getStatus_

        Get the Status for the workflow
        """
        return getattr(self.data.request, 'status', None)

    def setCampaign(self, campaign):
        """
        _setCampaign_

        Set the campaign to which this workflow belongs
        Optional
        """
        self.data.properties.campaign = campaign
        return

    def getCampaign(self):
        """
        _getCampaign_

        Get the campaign for the workflow
        """
        return getattr(self.data.properties, 'campaign', None)

    def setLFNBase(self, mergedLFNBase, unmergedLFNBase, runNumber=None):
        """
        _setLFNBase_

        Set the merged and unmerged base LFNs for all tasks.  Update all of the
        output LFNs to use them.
        """
        self.data.properties.mergedLFNBase = mergedLFNBase
        self.data.properties.unmergedLFNBase = unmergedLFNBase

        # set all child tasks lfn base.
        for task in self.taskIterator():
            task.setLFNBase(mergedLFNBase, unmergedLFNBase)
        self.updateLFNsAndDatasets(runNumber=runNumber)
        return

    def setMergeParameters(self, minSize, maxSize, maxEvents,
                           initialTask=None):
        """
        _setMergeParameters_

        Set the parameters for every merge task in the workload.  Also update
        the min merge size of every CMSSW step.
        """
        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            for stepName in task.listAllStepNames():
                stepHelper = task.getStepHelper(stepName)
                if stepHelper.stepType() == "StageOut" and stepHelper.minMergeSize() > 0:
                    stepHelper.setMinMergeSize(minSize, maxEvents)

            if task.taskType() == "Merge":
                task.setSplittingParameters(min_merge_size=minSize,
                                            max_merge_size=maxSize,
                                            max_merge_events=maxEvents)

            self.setMergeParameters(minSize, maxSize, maxEvents, task)

        return

    def setWorkQueueSplitPolicy(self, policyName, splitAlgo, splitArgs, **kwargs):
        """
        Sets the WorkQueue start policy.

        :param policyName: string with the policy name. Supported values should match  the
            WMCore/WorkQueue/Policy/Start module names (Dataset, Block, MonteCarlo or ResubmitBlock)
        :param splitAlgo: string with the job splitting algorithm name. Supported values should
            match the WMCore/JobSplitting module names.
        :param splitArgs: dictionary with job splitting arguments
        :param kwargs: dictionary with additional arguments that can be passed directly to
            the startPolicyArgs
        :return: None
        """
        SplitAlgoToStartPolicy = {"FileBased": ["NumberOfFiles"],
                                  "EventBased": ["NumberOfEvents",
                                                 "NumberOfEventsPerLumi"],
                                  "LumiBased": ["NumberOfLumis"],
                                  "Harvest": ["NumberOfRuns"],
                                  "EventAwareLumiBased": ["NumberOfEvents"]}
        SplitAlgoToArgMap = {"NumberOfFiles": "files_per_job",
                             "NumberOfEvents": "events_per_job",
                             "NumberOfLumis": "lumis_per_job",
                             "NumberOfRuns": "runs_per_job",
                             "NumberOfEventsPerLumi": "events_per_lumi"}
        startPolicyArgs = {'SplittingAlgo': splitAlgo}
        startPolicyArgs.update(kwargs)

        sliceTypes = SplitAlgoToStartPolicy.get(splitAlgo, ["NumberOfFiles"])
        sliceType = sliceTypes[0]
        sliceSize = splitArgs.get(SplitAlgoToArgMap[sliceType], 1)
        startPolicyArgs["SliceType"] = sliceType
        startPolicyArgs["SliceSize"] = sliceSize

        if len(sliceTypes) > 1:
            subSliceType = sliceTypes[1]
            subSliceSize = splitArgs.get(SplitAlgoToArgMap[subSliceType],
                                         sliceSize)
            startPolicyArgs["SubSliceType"] = subSliceType
            startPolicyArgs["SubSliceSize"] = subSliceSize

        self.setStartPolicy(policyName, **startPolicyArgs)
        self.setEndPolicy("SingleShot")
        return

    def setJobSplittingParameters(self, taskPath, splitAlgo, splitArgs, updateOnly=False):
        """
        _setJobSplittingParameters_

        Update the job splitting algorithm and arguments for the given task.
        """
        taskHelper = self.getTaskByPath(taskPath)
        if taskHelper == None:
            return

        if taskHelper.isTopOfTree():
            self.setWorkQueueSplitPolicy(self.startPolicy(), splitAlgo, splitArgs)

        # There are currently two merge algorithms in WMBS.  WMBSMergeBySize
        # will reassemble the parent file.  This is only necessary for
        # EventBased processing where we break up lumi sections.  Everything
        # else can use ParentlessMergeBySize which won't reassemble parents.
        # Everything defaults to ParentlessMergeBySize as it is much less load
        # on the database.
        minMergeSize = None
        maxMergeEvents = None
        for childTask in taskHelper.childTaskIterator():
            if childTask.taskType() == "Merge":
                if splitAlgo == "EventBased" and taskHelper.taskType() != "Production":
                    mergeAlgo = "WMBSMergeBySize"
                    for stepName in childTask.listAllStepNames():
                        stepHelper = childTask.getStepHelper(stepName)
                        if stepHelper.stepType() == "CMSSW":
                            stepCmsswHelper = stepHelper.getTypeHelper()
                            stepCmsswHelper.setSkipBadFiles(False)
                else:
                    mergeAlgo = "ParentlessMergeBySize"

                childSplitParams = childTask.jobSplittingParameters()
                minMergeSize = childSplitParams["min_merge_size"]
                maxMergeEvents = childSplitParams["max_merge_events"]
                if not updateOnly:
                    del childSplitParams["algorithm"]
                    del childSplitParams["siteWhitelist"]
                    del childSplitParams["siteBlacklist"]
                    childTask.setSplittingAlgorithm(mergeAlgo, **childSplitParams)
                else:
                    childTask.updateSplittingParameters(mergeAlgo, **childSplitParams)
        # Set the splitting algorithm for the task.  If the split algo is
        # EventBased, we need to disable straight to merge.  If this isn't an
        # EventBased algo we need to enable straight to merge. If straight
        # to merge is disabled then keep it that way.
        if not updateOnly:
            taskHelper.setSplittingAlgorithm(splitAlgo, **splitArgs)
        else:
            taskHelper.updateSplittingParameters(splitAlgo, **splitArgs)
        for stepName in taskHelper.listAllStepNames():
            stepHelper = taskHelper.getStepHelper(stepName)
            if stepHelper.stepType() == "StageOut":
                if splitAlgo != "EventBased" and stepHelper.minMergeSize() != -1 and minMergeSize:
                    stepHelper.setMinMergeSize(minMergeSize, maxMergeEvents)
                else:
                    stepHelper.disableStraightToMerge()
            if stepHelper.stepType() == "CMSSW" and splitAlgo == "WMBSMergeBySize" \
                    and stepHelper.getSkipBadFiles():
                stepHelper.setSkipBadFiles(False)

            if taskHelper.isTopOfTree() and taskHelper.taskType() == "Production" and stepName == "cmsRun1":
                # set it only for the first cmsRun in multi-steps tasks
                stepHelper.setEventsPerLumi(splitArgs.get("events_per_lumi", None))
        return

    def listJobSplittingParametersByTask(self, initialTask=None, performance=True):
        """
        _listJobSplittingParametersByTask_

        Create a dictionary that maps task names to job splitting parameters.
        """
        output = {}

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            taskName = task.getPathName()
            taskParams = task.jobSplittingParameters(performance)
            del taskParams["siteWhitelist"]
            del taskParams["siteBlacklist"]
            output[taskName] = taskParams
            output[taskName]["type"] = task.taskType()
            output.update(self.listJobSplittingParametersByTask(task, performance))

        return output

    def listInputDatasets(self):
        """
        _listInputDatasets_

        List all the input datasets in the workload
        """
        inputDatasets = []

        taskIterator = self.taskIterator()
        for task in taskIterator:
            path = task.getInputDatasetPath()
            if path:
                inputDatasets.append(path)

        return inputDatasets

    def listOutputDatasets(self, initialTask=None):
        """
        _listOutputDatasets_

        List the names of all the datasets produced by this workflow.
        """
        outputDatasets = []

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            for stepName in task.listAllStepNames():
                stepHelper = task.getStepHelper(stepName)

                if not getattr(stepHelper.data.output, "keep", True):
                    continue

                if stepHelper.stepType() == "CMSSW":
                    for outputModuleName in stepHelper.listOutputModules():
                        # Only consider non-transient output
                        outputModule = stepHelper.getOutputModule(outputModuleName)
                        if getattr(outputModule, "transient", False):
                            continue
                        outputDataset = "/%s/%s/%s" % (outputModule.primaryDataset,
                                                       outputModule.processedDataset,
                                                       outputModule.dataTier)
                        outputDatasets.append(outputDataset)

            moreDatasets = self.listOutputDatasets(task)
            outputDatasets.extend(moreDatasets)

        return outputDatasets

    def listAllOutputModulesLFNBases(self, initialTask=None, onlyUnmerged=True):
        """
        _listAllOutputModulesLFNBases_

        List all output LFN bases defined in this workload object.
        """
        listLFNBases = set()
        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            for stepName in task.listAllStepNames():
                outModule = task.getOutputModulesForStep(stepName)
                for module in viewvalues(outModule.dictionary_()):
                    lfnBase = getattr(module, "lfnBase", "")
                    if not onlyUnmerged and lfnBase:
                        listLFNBases.add(lfnBase)
                    elif lfnBase.startswith('/store/unmerged'):
                        listLFNBases.add(lfnBase)
            # recursively go through all the tasks
            listLFNBases.update(self.listAllOutputModulesLFNBases(task, onlyUnmerged))

        return list(listLFNBases)

    def listPileupDatasets(self, initialTask=None):
        """
        _listPileUpDataset_

        Returns a dictionary with all the required pile-up datasets
        in this workload and their associated dbs url as the key
        """
        pileupDatasets = {}

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            for stepName in task.listAllStepNames():
                stepHelper = task.getStepHelper(stepName)
                if stepHelper.stepType() == "CMSSW":
                    pileupSection = stepHelper.getPileup()
                    if pileupSection is None:
                        continue
                    dbsUrl = stepHelper.data.dbsUrl
                    if dbsUrl not in pileupDatasets:
                        pileupDatasets[dbsUrl] = set()
                    for pileupType in pileupSection.listSections_():
                        datasets = getattr(getattr(stepHelper.data.pileup, pileupType), "dataset")
                        pileupDatasets[dbsUrl].update(datasets)

            pileupDatasets.update(self.listPileupDatasets(task))

        return pileupDatasets

    def listOutputProducingTasks(self, initialTask=None):
        """
        _listOutputProducingTasks_

        List the paths to any task capable of producing merged output
        """
        taskList = []

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            for stepName in task.listAllStepNames():
                stepHelper = task.getStepHelper(stepName)
                if not getattr(stepHelper.data.output, "keep", True):
                    continue

                if stepHelper.stepType() == "CMSSW":
                    if stepHelper.listOutputModules():
                        taskList.append(task.getPathName())
                        break

            taskList.extend(self.listOutputProducingTasks(task))

        return taskList

    def setSubscriptionInformation(self, initialTask=None, custodialSites=None,
                                   nonCustodialSites=None,
                                   priority="Low", primaryDataset=None,
                                   useSkim=False, isSkim=False,
                                   dataTier=None, deleteFromSource=False,
                                   datasetLifetime=None):
        """
        _setSubscriptionInformation_

        Set the given subscription information for all datasets
        in the workload that match the given primaryDataset (if any)
        """

        if custodialSites and not isinstance(custodialSites, list):
            custodialSites = [custodialSites]
        if nonCustodialSites and not isinstance(nonCustodialSites, list):
            nonCustodialSites = [nonCustodialSites]

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            task.setSubscriptionInformation(custodialSites, nonCustodialSites,
                                            priority, primaryDataset,
                                            useSkim, isSkim,
                                            dataTier, deleteFromSource,
                                            datasetLifetime)
            self.setSubscriptionInformation(task,
                                            custodialSites, nonCustodialSites,
                                            priority, primaryDataset,
                                            useSkim, isSkim,
                                            dataTier, deleteFromSource,
                                            datasetLifetime)

        return

    def getSubscriptionInformation(self, initialTask=None):
        """
        _getSubscriptionInformation_

        Get the subscription information for the whole workload, this is given by
        dataset and aggregated according to the information from each individual task
        See WMTask.WMTaskHelper.getSubscriptionInformation for the output structure
        """
        subInfo = {}

        # Add site lists without duplicates
        extendWithoutDups = lambda x, y: x + list(set(y) - set(x))
        # Choose the lowest priority
        solvePrioConflicts = lambda x, y: y if x == "High" or y == "Low" else x
        # Always choose a logical AND
        solveDelConflicts = lambda x, y: x and y

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
            subInfo = initialTask.getSubscriptionInformation()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            taskSubInfo = self.getSubscriptionInformation(task)
            for dataset in taskSubInfo:
                if dataset in subInfo:
                    subInfo[dataset]["CustodialSites"] = extendWithoutDups(taskSubInfo[dataset]["CustodialSites"],
                                                                           subInfo[dataset]["CustodialSites"])
                    subInfo[dataset]["NonCustodialSites"] = extendWithoutDups(taskSubInfo[dataset]["NonCustodialSites"],
                                                                              subInfo[dataset]["NonCustodialSites"])
                    subInfo[dataset]["Priority"] = solvePrioConflicts(taskSubInfo[dataset]["Priority"],
                                                                      subInfo[dataset]["Priority"])
                    subInfo[dataset]["DeleteFromSource"] = solveDelConflicts(taskSubInfo[dataset]["DeleteFromSource"],
                                                                             subInfo[dataset]["DeleteFromSource"])
                else:
                    subInfo[dataset] = taskSubInfo[dataset]
                subInfo[dataset]["CustodialSites"] = list(
                    set(subInfo[dataset]["CustodialSites"]) - set(subInfo[dataset]["NonCustodialSites"]))

        return subInfo

    def getWorkloadOverrides(self):
        """
        _getWorkloadOverrides_

        Get the overrides config section
        of this workload, creates it if it doesn't exist
        """
        return self.data.section_('overrides')

    def setWorkloadOverrides(self, overrides):
        """
        _setWorkloadOverrides_

        Set the override parameters for all logArch steps
        in all tasks.
        """
        if overrides:
            for task in self.getAllTasks():
                for stepName in task.listAllStepNames():
                    stepHelper = task.getStepHelper(stepName)
                    if stepHelper.stepType() == "LogArchive":
                        for key, value in viewitems(overrides):
                            stepHelper.addOverride(key, value)
            # save it at workload level as well
            for key, value in viewitems(overrides):
                setattr(self.data.overrides, key, value)

        return

    def setBlockCloseSettings(self, blockCloseMaxWaitTime,
                              blockCloseMaxFiles, blockCloseMaxEvents,
                              blockCloseMaxSize):
        """
        _setBlockCloseSettings_

        Set the parameters that define when a block should be closed
        for this workload, they should all be defined so it is a single call
        """
        self.data.properties.blockCloseMaxWaitTime = blockCloseMaxWaitTime
        self.data.properties.blockCloseMaxFiles = blockCloseMaxFiles
        self.data.properties.blockCloseMaxEvents = blockCloseMaxEvents
        self.data.properties.blockCloseMaxSize = blockCloseMaxSize

    def getBlockCloseMaxWaitTime(self):
        """
        _getBlockCloseMaxWaitTime_

        Return the amount of time that a block should stay open
        for this workload before closing it in DBS
        """

        return getattr(self.data.properties, 'blockCloseMaxWaitTime', 66400)

    def getBlockCloseMaxSize(self):
        """
        _getBlockCloseMaxSize_

        Return the maximum size that a block from this workload should have
        """

        return getattr(self.data.properties, 'blockCloseMaxSize', 5000000000000)

    def getBlockCloseMaxEvents(self):
        """
        _blockCloseMaxEvents_

        Return the maximum number of events that a block from this workload
        should have
        """

        return getattr(self.data.properties, 'blockCloseMaxEvents', 25000000)

    def getBlockCloseMaxFiles(self):
        """
        _getBlockCloseMaxFiles_

        Return the maximum amount of files that a block from this workload
        should have
        """

        return getattr(self.data.properties, 'blockCloseMaxFiles', 500)

    def getUnmergedLFNBase(self):
        """
        _getUnmergedLFNBase_

        Get the unmerged LFN Base from properties
        """

        return getattr(self.data.properties, 'unmergedLFNBase', None)

    def getMergedLFNBase(self):
        """
        _getMergedLFNBase_

        Get the merged LFN Base from properties
        """

        return getattr(self.data.properties, 'mergedLFNBase', None)

    def getLFNBases(self):
        """
        _getLFNBases_

        Retrieve the LFN bases.  They are returned as a tuple with the merged
        LFN base first, followed by the unmerged LFN base.
        """
        return self.getMergedLFNBase(), self.getUnmergedLFNBase()

    def setRetryPolicy(self):
        """
        _setRetryPolicy_

        """
        pass

    def truncate(self, newWorkloadName, initialTaskPath, serverUrl,
                 databaseName, collectionName=None):
        """
        _truncate_

        Truncate a workflow so that it can be used for resubmission.  This will
        rename the workflow and set the task in the intitialTaskPath parameter
        to be the top level task.  This modifies the workflow in place.
        The input collection name can be specified otherwise it will default to
        the old workload name.
        """
        if not collectionName:
            collectionName = self.name()

        allTaskPaths = self.listAllTaskPathNames()
        newTopLevelTask = self.getTaskByPath(initialTaskPath)
        newTopLevelTask.addInputACDC(serverUrl, databaseName, collectionName,
                                     initialTaskPath)
        newTopLevelTask.setInputStep(None)
        workloadOwner = self.getOwner()
        self.setInitialJobCount(self.getInitialJobCount() + 10000000)
        newTopLevelTask.setSplittingParameters(collectionName=collectionName,
                                               filesetName=initialTaskPath,
                                               couchURL=serverUrl,
                                               couchDB=databaseName,
                                               owner=workloadOwner["name"],
                                               group=workloadOwner["group"],
                                               initial_lfn_counter=self.getInitialJobCount())

        for taskPath in allTaskPaths:
            if not taskPath.startswith(initialTaskPath) or taskPath == initialTaskPath:
                taskName = taskPath.split("/")[-1]
                if hasattr(self.data.tasks, taskName):
                    delattr(self.data.tasks, taskName)
                if taskName in self.data.tasks.tasklist:
                    self.data.tasks.tasklist.remove(taskName)

        self.setName(newWorkloadName)
        self.addTask(newTopLevelTask)
        newTopLevelTask.setTopOfTree()

        self.setWorkQueueSplitPolicy("ResubmitBlock",
                                     newTopLevelTask.jobSplittingAlgorithm(),
                                     newTopLevelTask.jobSplittingParameters())

        def adjustPathsForTask(initialTask, parentPath):
            """
            _adjustPathsForTask_

            Given an initial task and the path for that task set the path
            correctly for all of the children tasks.
            """
            for childTask in initialTask.childTaskIterator():
                childTask.setPathName("%s/%s" % (parentPath, childTask.name()))
                inputStep = childTask.getInputStep()
                if inputStep != None:
                    inputStep = inputStep.replace(parentPath, "/" + newWorkloadName)
                    childTask.setInputStep(inputStep)

                adjustPathsForTask(childTask, childTask.getPathName())

            return

        adjustPathsForTask(newTopLevelTask, "/%s/%s" % (newWorkloadName,
                                                        newTopLevelTask.name()))
        return

    def ignoreOutputModules(self, badModules, initialTask=None):
        """
        _ignoreOutputModules_

        If there is a list of ignored output modules the following must be done:
        - Trim the workload tree so that no task that depends on the merged output of the ignored modules
          exists in the tree, also eliminate the merge task for such modules
        - Add flags to make the runtime code ignore the files from this module so they are not
          staged out
        """

        if not badModules:
            return

        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            # Find the children tasks that have a bad output module as
            # input, disown them. Can't delete them on the spot, save the names in a list
            childTasksToDelete = []
            for childTask in task.childTaskIterator():
                taskInput = childTask.inputReference()
                inputOutputModule = getattr(taskInput, "outputModule", None)
                if inputOutputModule in badModules:
                    childTasksToDelete.append(childTask.name())

            # Now delete
            for childTaskName in childTasksToDelete:
                task.deleteChild(childTaskName)

            if childTasksToDelete:
                # Tell any CMSSW step to ignore the output modules
                for stepName in task.listAllStepNames():
                    stepHelper = task.getStepHelper(stepName)
                    if stepHelper.stepType() == "CMSSW":
                        stepHelper.setIgnoredOutputModules(badModules)
            # Go deeper in the tree
            self.ignoreOutputModules(badModules, task)

        return

    def setCMSSWVersions(self, cmsswVersion=None, globalTag=None,
                         scramArch=None, initialTask=None):
        """
        _setCMSSWVersions_

        Set the CMSSW version and the global tag for all CMSSW steps in the
        workload.
        """
        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            for stepName in task.listAllStepNames():
                stepHelper = task.getStepHelper(stepName)

                if stepHelper.stepType() == "CMSSW":
                    if cmsswVersion != None:
                        if scramArch != None:
                            stepHelper.cmsswSetup(cmsswVersion=cmsswVersion,
                                                  scramArch=scramArch)
                        else:
                            stepHelper.cmsswSetup(cmsswVersion=cmsswVersion)

                    if globalTag != None:
                        stepHelper.setGlobalTag(globalTag)

            self.setCMSSWVersions(cmsswVersion, globalTag, scramArch, task)

        return

    def getCMSSWVersions(self, initialTask=None):
        """
        _getCMSSWVersions_

        Return a list of all CMSSW releases being used in this workload.
        """
        versions = set()
        if initialTask:
            taskIterator = initialTask.childTaskIterator()
        else:
            taskIterator = self.taskIterator()

        for task in taskIterator:
            for stepName in task.listAllStepNames():
                stepHelper = task.getStepHelper(stepName)
                if stepHelper.stepType() == "CMSSW":
                    versions.add(stepHelper.getCMSSWVersion())
            versions.update(self.getCMSSWVersions(task))

        return list(versions)

    def generateWorkloadSummary(self):
        """
        _generateWorkloadSummary_

        Generates a dictionary with the following information:
        task paths
        ACDC
        input datasets
        output datasets

        Intended for use in putting WMSpec info into couch
        """
        summary = {'tasks': [],
                   'ACDC': {"collection": None, "filesets": {}},
                   'input': [],
                   'output': [],
                   'owner': {},
                   }

        summary['tasks'] = self.listAllTaskPathNames()
        summary['output'] = self.listOutputDatasets()
        summary['input'] = self.listInputDatasets()
        summary['owner'] = self.data.owner.dictionary_()
        summary['performance'] = {}
        for t in summary['tasks']:
            summary['performance'][t] = {}

        return summary

    def setupPerformanceMonitoring(self, softTimeout, gracePeriod):
        """
        _setupPerformanceMonitoring_

        Setups performance monitors for all tasks in the workflow
        """
        for task in self.getAllTasks():
            task.setPerformanceMonitor(softTimeout=softTimeout, gracePeriod=gracePeriod)

        return

    def listAllCMSSWConfigCacheIDs(self):
        """
        _listAllCMSSWConfigCacheIDs_

        Go through each task and check to see if we have a configCacheID
        """
        result = []
        for t in self.taskIterator():
            result.extend(t.getConfigCacheIDs())
        return result

    def setTrustLocationFlag(self, inputFlag=False, pileupFlag=False):
        """
        Set the input and the pileup flags in the top level tasks
        indicating that site lists should be used as location data

        The input data flag has to be set only for top level tasks, otherwise
        it affects where secondary jobs are meant to run.
        The pileup flag has to be set for all the tasks in the workload.

        Validate these parameters to make sure they are only set for workflows
        that require those type of input datasets (ACDCs are not validated)
        """
        isACDCWorkflow = False
        for task in self.taskIterator():
            if task.getInputACDC():
                isACDCWorkflow = True

        if inputFlag is True and isACDCWorkflow is False and not self.listInputDatasets():
            msg = "Setting TrustSitelists=True for workflows without input dataset is forbidden!"
            raise RuntimeError(msg)
        if pileupFlag is True and isACDCWorkflow is False and not self.listPileupDatasets():
            msg = "Setting TrustPUSitelists=True for workflows without pileup dataset is forbidden!"
            raise RuntimeError(msg)
        for task in self.getAllTasks(cpuOnly=True):
            if task.isTopOfTree():
                task.setTrustSitelists(inputFlag, pileupFlag)
            else:
                task.setTrustSitelists(False, pileupFlag)

        return

    def getTrustLocationFlag(self):
        """
        _getTrustLocationFlag_

        Get a tuple with the inputFlag and the pileupFlag values from
        the *top level* tasks that indicates whether the site lists should
        be trusted as the location for input and/or for the pileup data.
        """
        for task in self.getTopLevelTask():
            return task.getTrustSitelists()
        return {'trustlists': False, 'trustPUlists': False}

    def validateArgumentForAssignment(self, schema):
        specClass = loadSpecClassByType(self.getRequestType())
        argumentDefinition = specClass.getWorkloadAssignArgs()
        validateArgumentsUpdate(schema, argumentDefinition)
        return

    def validateSiteListsUpdate(self, arguments):
        """
        Validate a dictionary of workflow arguments for possible  Site lists update
        It must catch any eventual conflict between the currently existing site lists
        and the ones provided by the user, presuming that the change would happen by
        fully substituting an already existing site list at the workflow if provided
        with the arguments here.
        """
        # NOTE: We have 3 different use cases to validate for siteLists conflicts:
        #       * A change to both SiteWhitelist and SiteBlacklist
        #       * A change only to SiteWhitelist
        #       * A change only to SiteBlacklist
        if "SiteWhitelist" in arguments and "SiteBlacklist" in arguments:
            validateSiteLists(arguments)
            return
        fullSiteWhitelist = set()
        fullSiteBlacklist = set()
        if "SiteWhitelist" in arguments and "SiteBlacklist" not in arguments:
            fullSiteWhitelist = set(makeList(arguments["SiteWhitelist"]))
            fullSiteBlacklist = set(self.getSiteBlacklist())
        if "SiteBlacklist" in arguments and "SiteWhitelist" not in arguments:
            fullSiteWhitelist = set(self.getSiteWhitelist())
            fullSiteBlacklist = set(makeList(arguments["SiteBlacklist"]))
        siteConflicts = fullSiteWhitelist & fullSiteBlacklist
        if siteConflicts:
            msg = "Validation of Site Lists for update failed due to conflicts with existing Site Lists. "
            msg += f"A site can only be black listed or whitelisted. Conflicting sites: {list(siteConflicts)}"
            raise WMSpecFactoryException(msg)
        return

    def validateArgumentsPartialUpdate(self, arguments):
        """
        Validates the provided parameters schema for workflow arguments update, using
        the arguments definitions for assignment as provided at StdBase.getWorkloadAssignArgs
        :param arguments: Workflow arguments schema to be validated - Must be a properly
                          defined dictionary of {arg: value} pairs.
        :return:          Nothing. Raises proper exceptions when argument validation fails

        NOTE: In order to avoid full schema validation and enforcing mandatory arguments,
              we set the optionKey argument for this call to NONE. This way it is ignored
              during the next step of the validation process (namely at
              WMWorkloadTools._validateArgumentOptions), and all of the so provided
              arguments in the schema are considered optional, but nonetheless they
              still go through the full value validation process.
        """
        specClass = loadSpecClassByType(self.getRequestType())
        argumentDefinition = specClass.getWorkloadAssignArgs()
        validateUnknownArgs(arguments, argumentDefinition)
        _validateArgumentOptions(arguments, argumentDefinition, optionKey=None)
        self.validateSiteListsUpdate(arguments)
        return

    def updateArguments(self, kwargs):
        """
        set up all the argument related to assigning request.
        args are validated before update.
        assignment is common for all different types spec.

        Input data should have been validated already using
        validateArgumentForAssignment.
        """
        specClass = loadSpecClassByType(self.getRequestType())
        argumentDefinition = specClass.getWorkloadAssignArgs()
        setAssignArgumentsWithDefault(kwargs, argumentDefinition)

        if kwargs.get('RequestPriority') is not None and kwargs['RequestPriority'] != self.priority():
            self.setPriority(kwargs['RequestPriority'])
        else:
            # if it's the same, pop it out to avoid priority transition update
            kwargs.pop("RequestPriority", None)

        self.setWorkloadOverrides(kwargs["Override"])
        self.setSiteWhitelist(kwargs["SiteWhitelist"])
        self.setSiteBlacklist(kwargs["SiteBlacklist"])
        self.setTrustLocationFlag(inputFlag=strToBool(kwargs["TrustSitelists"]),
                                  pileupFlag=strToBool(kwargs["TrustPUSitelists"]))

        self.setMergeParameters(int(kwargs["MinMergeSize"]),
                                int(kwargs["MaxMergeSize"]),
                                int(kwargs["MaxMergeEvents"]))

        # FIXME not validated
        if kwargs.get("MergedLFNBase") and kwargs.get("UnmergedLFNBase"):
            self.setLFNBase(kwargs["MergedLFNBase"], kwargs["UnmergedLFNBase"])
        # Set ProcessingVersion and AcquisitionEra, which could be json encoded dicts
        # it should be processed once LFNBase are set
        if kwargs.get("AcquisitionEra") is not None:
            self.setAcquisitionEra(kwargs["AcquisitionEra"])
        if kwargs.get("ProcessingString") is not None:
            self.setProcessingString(kwargs["ProcessingString"])
        if kwargs.get("ProcessingVersion") is not None:
            self.setProcessingVersion(kwargs["ProcessingVersion"])

        self.setupPerformanceMonitoring(softTimeout=kwargs["SoftTimeout"],
                                        gracePeriod=kwargs["GracePeriod"])

        # Check whether we should check location for the data
        self.setAllowOpportunistic(allowOpport=strToBool(kwargs["AllowOpportunistic"]))

        # Block closing information
        self.setBlockCloseSettings(kwargs["BlockCloseMaxWaitTime"],
                                   kwargs["BlockCloseMaxFiles"],
                                   kwargs["BlockCloseMaxEvents"],
                                   kwargs["BlockCloseMaxSize"])

        self.setDashboardActivity(kwargs["Dashboard"])

        if kwargs.get("Memory") is not None:
            self.setMemory(kwargs.get("Memory"))
        if kwargs.get("Multicore") is not None:
            self.setCoresAndStreams(kwargs.get("Multicore"), kwargs.get("EventStreams"))

        # MUST be set after AcqEra/ProcStr/ProcVer
        if self.getRequestType() == "StepChain":
            self.setStepProperties(kwargs)
            self.updateStepParentageMap()
        elif self.getRequestType() == "TaskChain":
            # TODO: need to define proper task form maybe kwargs['Tasks']?
            self.setTaskProperties(kwargs)
            self.updateTaskParentageMap()

        # Since it lists the output datasets, it has to be done in the very end
        # Set phedex subscription information
        if kwargs.get("CustodialSites") or kwargs.get("NonCustodialSites"):
            self.setSubscriptionInformation(custodialSites=kwargs["CustodialSites"],
                                            nonCustodialSites=kwargs["NonCustodialSites"],
                                            priority=kwargs["SubscriptionPriority"],
                                            deleteFromSource=kwargs["DeleteFromSource"])

        return

    def loadSpecFromCouch(self, couchurl, requestName):
        """
        This depends on PersitencyHelper.py saveCouch (That method better be decomposed)
        """
        return self.load("%s/%s/spec" % (couchurl, requestName))

    def setTaskPropertiesFromWorkload(self):
        """
        set task properties inherits from workload properties
        This is need to be called at the end of the buildWorkload function
        after all the tasks are added.
        It sets acquisitionEra, processingVersion, processingString,
        since those values are needed to be set for all the tasks in the workload
        TODO: need to force to call this function after task is added instead of
              rely on coder's won't forget to call this at the end of
              self.buildWorkload()
        """
        self.setAcquisitionEra(self.getAcquisitionEra())
        self.setProcessingVersion(self.getProcessingVersion())
        self.setProcessingString(self.getProcessingString())
        self.setLumiList(self.getLumiList())
        self.setPrepID(self.getPrepID())
        return


class WMWorkload(ConfigSection):
    """
    _WMWorkload_

    Request container

    """

    def __init__(self, name="test"):
        ConfigSection.__init__(self, name)
        self.objectType = self.__class__.__name__
        #  //persistent data
        # //
        # //
        self.section_("persistency")
        self.persistency.specUrl = None
        #  //
        # // request related information
        # //
        self.section_("request")
        self.request.priority = None  # what should be the default value
        self.request.status = None
        #  //
        # // owner related information
        # //
        self.section_("owner")

        #  //
        # // Policies applied to this workload by the processing system
        # //
        self.section_("policies")
        self.policies.section_("start")
        self.policies.section_("end")
        self.policies.start.policyName = None
        self.policies.end.policyName = None

        #  //
        # // properties of the Workload and all tasks there-in
        # //
        self.section_("properties")
        self.properties.unmergedLFNBase = "/store/unmerged"
        self.properties.mergedLFNBase = "/store/data"
        self.properties.dashboardActivity = None
        self.properties.blockCloseMaxWaitTime = 66400
        self.properties.blockCloseMaxSize = 5000000000000
        self.properties.blockCloseMaxFiles = 500
        self.properties.blockCloseMaxEvents = 250000000
        self.properties.prepID = None

        # Overrides for this workload
        self.section_("overrides")

        #  //
        # // tasks
        # //
        self.section_("tasks")
        self.tasks.tasklist = []

        #  workload spec type
        self.requestType = ""
        self.dbsUrl = None

        self.sandbox = None
        self.initialJobCount = 0


def newWorkload(workloadName):
    """
    _newWorkload_

    Util method to create a new WMWorkload and wrap it in a helper

    """
    return WMWorkloadHelper(WMWorkload(workloadName))



"""
Given a pile up dataset, pull the information required to cache the list
of pileup files in the job sandbox for the dataset.

"""
import datetime
import os
import hashlib
import shutil
import time
import logging
from json import JSONEncoder
import WMCore.WMSpec.WMStep as WMStep
from Utils.Patterns import getDomainName
from Utils.Utilities import encodeUnicodeToBytes
from WMCore.Services.DBS.DBSReader import DBSReader
from WMCore.Services.MSUtils.MSUtils import getPileupDocs
from WMCore.Services.Rucio.Rucio import Rucio
from WMCore.WMSpec.Steps.Fetchers.FetcherInterface import FetcherInterface


class PileupFetcher(FetcherInterface):
    """
    Pull dataset block/SE : LFN list from DBS for the
    pileup datasets required by the steps in the job.

    Save these maps as files in the sandbox

    """
    def __init__(self):
        """
        Prepare module setup
        """
        super(PileupFetcher, self).__init__()
        self.rucioAcct = "wmcore_pileup"
        self.rucio = None

    def _queryDbsAndGetPileupConfig(self, stepHelper, dbsReader):
        """
        Method iterates over components of the pileup configuration input
        and queries DBS for valid files in the dataset, plus some extra
        information about each file.

        Information is organized at block level, listing all its files,
        number of events in the block, and its data location (to be resolved
        by a different method using either PhEDEx or Rucio), such as:

        {"pileupTypeA": {"BlockA": {"FileList": [], "PhEDExNodeNames": [], "NumberOfEvents": 123},
                         "BlockB": {"FileList": [], "PhEDExNodeName": []}, ....}
        """
        resultDict = {}
        # first, figure out which instance of MSPileup and Rucio to use
        pileupInstance = getDomainName(dbsReader.dbsURL)
        msPileupUrl = f"https://{pileupInstance}.cern.ch/ms-pileup/data/pileup"
        # FIXME: this juggling with Rucio is tough! We can get away without it,
        # but for that we would have to use testbed MSPileup against Prod Rucio
        if pileupInstance == "cmsweb-prod" or pileupInstance == "cmsweb":
            rucioAuthUrl, rucioUrl = "cms-rucio-auth", "cms-rucio"
        else:
            rucioAuthUrl, rucioUrl = "cms-rucio-auth-int", "cms-rucio-int"
        # initialize Rucio here to avoid this authentication on T0-WMAgent
        self.rucio = Rucio(self.rucioAcct,
                           authUrl=f"https://{rucioAuthUrl}.cern.ch",
                           hostUrl=f"http://{rucioUrl}.cern.ch")

        # iterate over input pileup types (e.g. "cosmics", "minbias")
        for pileupType in stepHelper.data.pileup.listSections_():
            # the format here is: step.data.pileup.cosmics.dataset = [/some/data/set]
            datasets = getattr(getattr(stepHelper.data.pileup, pileupType), "dataset")
            # each dataset input can generally be a list, iterate over dataset names
            blockDict = {}
            for dataset in datasets:
                # using the original dataset, resolve blocks, files and number of events with DBS
                fCounter = 0
                for fileInfo in dbsReader.getFileListByDataset(dataset=dataset, detail=True):
                    blockDict.setdefault(fileInfo['block_name'], {'FileList': [],
                                                                  'NumberOfEvents': 0,
                                                                  'PhEDExNodeNames': []})
                    blockDict[fileInfo['block_name']]['FileList'].append(fileInfo['logical_file_name'])
                    blockDict[fileInfo['block_name']]['NumberOfEvents'] += fileInfo['event_count']
                    fCounter += 1

                logging.info(f"Found {len(blockDict)} blocks in DBS for dataset {dataset} with {fCounter} files")
                self._getDatasetLocation(dataset, blockDict, msPileupUrl)

            resultDict[pileupType] = blockDict
        return resultDict

    def _getDatasetLocation(self, dset, blockDict, msPileupUrl):
        """
        Given a dataset name, query PhEDEx or Rucio and resolve the block location
        :param dset: string with the dataset name
        :param blockDict: dictionary with DBS summary info
        :param msPileupUrl: string with the MSPileup url
        :return: update blockDict in place
        """
        # fetch the pileup configuration from MSPileup
        try:
            queryDict = {'query': {'pileupName': dset},
                                   'filters': ['pileupName', 'customName', 'containerFraction', 'currentRSEs']}
            doc = getPileupDocs(msPileupUrl, queryDict, method='POST')[0]
            msg = f'Pileup dataset {doc["pileupName"]} with:\n\tcustom name: {doc["customName"]},'
            msg += f'\n\tcurrent RSEs: {doc["currentRSEs"]}\n\tand container fraction: {doc["containerFraction"]}'
            logging.info(msg)
        except Exception as ex:
            logging.error(f'Error querying MSPileup for dataset {dset}. Details: {str(ex)}')
            raise ex

        # custom dataset name means there was a container fraction change, use different scope
        puScope = 'cms'
        if doc["customName"]:
            dset = doc["customName"]
            puScope = 'group.wmcore'

        blockReplicas = self.rucio.getBlocksInContainer(container=dset, scope=puScope)
        logging.info(f"Found {len(blockReplicas)} blocks in container {dset} for scope {puScope}")

        # Finally, update blocks present in Rucio with the MSPileup currentRSEs.
        # Blocks not present in Rucio - hence only in DBS - are meant to be removed.
        for blockName in list(blockDict):
            if blockName not in blockReplicas:
                logging.warning(f"Block {blockName} present in DBS but not in Rucio. Removing it.")
                blockDict.pop(blockName)
            else:
                blockDict[blockName]['PhEDExNodeNames'] = doc["currentRSEs"]
        logging.info(f"Final pileup dataset {dset} has a total of {len(blockDict)} blocks.")

    def _getCacheFilePath(self, stepHelper):

        fileName = ""
        for pileupType in stepHelper.data.pileup.listSections_():
            datasets = getattr(getattr(stepHelper.data.pileup, pileupType), "dataset")
            fileName += "_".join(datasets)
        # TODO cache is not very effective if the dataset combination is different between workflow
        cacheHash = hashlib.sha1(encodeUnicodeToBytes(fileName)).hexdigest()
        cacheFile = "%s/pileupconf-%s.json" % (self.cacheDirectory(), cacheHash)
        return cacheFile

    def _getStepFilePath(self, stepHelper):
        stepPath = "%s/%s" % (self.workingDirectory(), stepHelper.name())
        fileName = "%s/%s" % (stepPath, "pileupconf.json")

        return fileName

    def _writeFile(self, filePath, jsonPU):

        directory = filePath.rsplit('/', 1)[0]

        if not os.path.exists(directory):
            os.mkdir(directory)
        try:
            with open(filePath, 'w') as f:
                f.write(jsonPU)
        except IOError:
            m = "Could not save pileup JSON configuration file: '%s'" % filePath
            raise RuntimeError(m)

    def _copyFile(self, src, dest):

        directory = dest.rsplit('/', 1)[0]

        if not os.path.exists(directory):
            os.mkdir(directory)
        shutil.copyfile(src, dest)

    def _isCacheExpired(self, cacheFilePath, delta=24):
        """Is the cache expired? At delta hours (default 24) in the future.
        """
        # cache can either be a file name or an already opened file object

        if not os.path.exists(cacheFilePath):
            return True

        delta = datetime.timedelta(hours=delta)
        t = datetime.datetime.now() - delta
        # cache file mtime has been set to cache expiry time
        if os.path.getmtime(cacheFilePath) < time.mktime(t.timetuple()):
            return True

        return False

    def _isCacheValid(self, stepHelper):
        """
        Check whether cache is exits
        TODO: if the cacheDirectory is not inside the Sandbox it should not autormatically deleted.
              We can add cache refresh policy here
        """
        cacheFile = self._getCacheFilePath(stepHelper)

        if not self._isCacheExpired(cacheFile, delta=0.5) and os.path.getsize(cacheFile) > 0:
            # if file already exist don't make a new dbs call and overwrite the file.
            # just return
            fileName = self._getStepFilePath(stepHelper)
            if not os.path.isfile(fileName) or os.path.getsize(fileName) != os.path.getsize(cacheFile):
                self._copyFile(cacheFile, fileName)
            return True
        else:
            return False

    def _saveFile(self, stepHelper, jsonPU):

        cacheFile = self._getCacheFilePath(stepHelper)
        self._writeFile(cacheFile, jsonPU)
        fileName = self._getStepFilePath(stepHelper)
        self._copyFile(cacheFile, fileName)

    def createPileupConfigFile(self, helper):
        """
        Stores pileup JSON configuration file in the working
        directory / sandbox.
        :param helper: WMStepHelper instance
        :return: None
        """
        if self._isCacheValid(helper):
            # if file already exist don't make a new dbs call and overwrite the file.
            # just return
            return

        encoder = JSONEncoder()
        # this should have been set in CMSSWStepHelper along with
        # the pileup configuration
        url = helper.data.dbsUrl
        dbsReader = DBSReader(url)

        configDict = self._queryDbsAndGetPileupConfig(helper, dbsReader)

        # create JSON and save into a file
        jsonPU = encoder.encode(configDict)
        self._saveFile(helper, jsonPU)

    def __call__(self, wmTask):
        """
        Method is called  when WorkQueue creates the sandbox for a job.
        Need to look at the pileup configuration in the spec and query dbs to
        determine the lfns for the files in the datasets and what sites they're
        located at (WQ creates the job sandbox).

        wmTask is instance of WMTask.WMTaskHelper

        """
        for step in wmTask.steps().nodeIterator():
            helper = WMStep.WMStepHelper(step)
            # returns e.g. instance of CMSSWHelper
            # doesn't seem to be necessary ... strangely (some inheritance involved?)
            # typeHelper = helper.getTypeHelper()
            if hasattr(helper.data, "pileup"):
                self.createPileupConfigFile(helper)



#!/usr/bin/env python
"""
LogDB provides functionality to post/search messages into LogDB.
https://github.com/dmwm/WMCore/issues/5705
"""
# futures
from builtins import object
from future.utils import viewitems

from future import standard_library
standard_library.install_aliases()

# standard modules
import logging
import re
import threading
from collections import defaultdict
from http.client import HTTPException

# project modules
from WMCore.Lexicon import splitCouchServiceURL
from WMCore.Services.LogDB.LogDBBackend import LogDBBackend


def getLogDBInstanceFromThread():
    """This function only gets to call when LogDB is instantiated before hand
       All the WMComponentWorkers instatntiate LogDB automatically
    """
    myThread = threading.currentThread()
    if not hasattr(myThread, "logdbClient") or not isinstance(myThread.logdbClient, LogDB):
        # logdb is not set do anything
        return None
    return myThread.logdbClient


class LogDB(object):
    """
    _LogDB_

    LogDB object - interface to LogDB functionality.
    """

    def __init__(self, url, identifier, logger=None, **kwds):
        self.logger = logger if logger else logging.getLogger()
        self.url = url if url else 'https://cmsweb.cern.ch/couchdb/wmstats_logdb'
        self.identifier = identifier if identifier else 'unknown'
        self.default_user = "HEARTBEAT"

        try:
            self.thread_name = kwds.pop('thread_name')
        except KeyError:
            self.thread_name = threading.currentThread().getName()

        self.user_pat = re.compile(r'^/[a-zA-Z][a-zA-Z0-9/\=\s()\']*\=[a-zA-Z0-9/\=\.\-_/#:\s\']*$')
        self.agent = 0 if self.user_pat.match(self.identifier) else 1
        couch_url, db_name = splitCouchServiceURL(self.url)
        self.backend = LogDBBackend(couch_url, db_name, identifier,
                                    self.thread_name, agent=self.agent, **kwds)
        self.logger.info(self)

    def __repr__(self):
        "Return representation for class"
        return "<LogDB(url=%s, identifier=%s, agent=%d)>" \
               % (self.url, self.identifier, self.agent)

    def post(self, request=None, msg="", mtype="comment"):
        """Post new entry into LogDB for given request"""
        res = 'post-error'
        try:
            if request is None:
                request = self.default_user
            if self.user_pat.match(self.identifier):
                res = self.backend.user_update(request, msg, mtype)
            else:
                res = self.backend.agent_update(request, msg, mtype)
        except HTTPException as ex:
            msg = "Failed to post doc to LogDB. Reason: %s, status: %s" % (ex.reason, ex.status)
            self.logger.error(msg)
        except Exception as exc:
            self.logger.error("LogDBBackend post API failed, error=%s", str(exc))
        self.logger.debug("LogDB post request, res=%s", res)
        return res

    def get(self, request=None, mtype=None):
        """Retrieve all entries from LogDB for given request"""
        res = []
        try:
            if request is None:
                request = self.default_user
            if self.user_pat.match(self.identifier):
                agent = False
            else:
                agent = True
            for row in self.backend.get(request, mtype, agent=agent).get('rows', []):
                request = row['doc']['request']
                identifier = row['doc']['identifier']
                thr = row['doc']['thr']
                mtype = row['doc']['type']
                for rec in row['doc']['messages']:
                    rec.update({'request': request, 'identifier': identifier, 'thr': thr, 'type': mtype})
                    res.append(rec)
        except HTTPException as ex:
            msg = "Failed to get doc from LogDB. Reason: %s, status: %s" % (ex.reason, ex.status)
            self.logger.error(msg)
            res = 'get-error'
        except Exception as exc:
            self.logger.error("LogDBBackend get API failed, error=%s", str(exc))
            res = 'get-error'
        self.logger.debug("LogDB get request, res=%s", res)
        return res

    def get_all_requests(self):
        """Retrieve all entries from LogDB for given request"""
        try:
            results = self.backend.get_all_requests()
            res = []
            for row in results['rows']:
                res.append(row["key"])
        except Exception as exc:
            self.logger.error("LogDBBackend get_all_requests API failed, error=%s", str(exc))
            res = 'get-error'
        self.logger.debug("LogDB get_all_requests request, res=%s", res)
        return res

    def delete(self, request=None, mtype=None, this_thread=False, agent=True):
        """
        Delete entry in LogDB for given request
        if mtype == None - delete all the log for that request
        mtype != None - only delete specified mtype
        """
        res = 'delete-error'
        try:
            if request is None:
                request = self.default_user
            res = self.backend.delete(request, mtype, this_thread, agent)
        except HTTPException as ex:
            msg = "Failed to delete doc in LogDB. Reason: %s, status: %s" % (ex.reason, ex.status)
            self.logger.error(msg)
        except Exception as ex:
            self.logger.error("LogDBBackend delete API failed, error=%s", str(ex))
        self.logger.debug("LogDB delete request, res=%s", res)
        return res

    def cleanup(self, thr, backend='local'):
        """Clean-up back-end LogDB"""
        docs = []
        try:
            docs = self.backend.cleanup(thr)
        except Exception as exc:
            self.logger.error('LogDBBackend cleanup API failed, backend=%s, error=%s', backend, str(exc))
        return docs

    def heartbeat_report(self):
        report = defaultdict(dict)
        if self.user_pat.match(self.identifier):
            self.logger.error("User %s: doesn't allow this function", self.identifier)
            return report

        for row in self.backend.get(self.default_user, None, agent=True).get('rows', []):
            identifier = row['doc']['identifier']
            # wmstats thread can run in multiple boxed.
            if self.identifier == identifier or identifier.startswith(self.identifier):
                # this will handle wmstats DataCacheUpdate thread with multiple machine
                postfix = identifier.replace(self.identifier, "")
                thr = "%s%s" % (row['doc']['thr'], postfix)
                mtype = row['doc']['type']
                ts = row['doc']['messages'][-1]['ts']
                msg = row['doc']['messages'][-1]['msg']
                if (thr in report) and ('ts' in report[thr]) and ts <= report[thr]['ts']:
                    continue
                else:
                    report[thr]['type'] = mtype
                    report[thr]['msg'] = msg
                    report[thr]['ts'] = ts
        return report

    def _append_down_component_detail(self, report, thr, msg, ts=0, state="error"):
        report['down_components'].append(thr)
        detail = {'name': thr, 'worker_name': thr, 'state': state,
                  'last_error': ts, 'error_message': msg,
                  'pid': 'N/A'}
        report['down_component_detail'].append(detail)
        return

    def wmstats_down_components_report(self, thread_list):
        report = {}
        report['down_components'] = []
        report['down_component_detail'] = []

        hbinfo = self.heartbeat_report()
        for thr in thread_list:
            # skip DataCacheUpdate thread. It will have multiple with post fix.
            # i.e. DataCacheUpdate-vocms111
            # TODO, need a better way to check
            if thr != "DataCacheUpdate" and thr not in hbinfo:
                self._append_down_component_detail(report, thr, "Thread not running")

        for thr, info in viewitems(hbinfo):
            if info['type'] == 'agent-error':
                self._append_down_component_detail(report, thr, info['msg'], info['ts'])
        return report



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : LogDBReport.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: LogDB report class to represent LogDB messages
"""
from __future__ import print_function

from builtins import range
from builtins import object

class LogDBReport(object):
    """LogDBReport class to represent LogDB messages"""
    def __init__(self, logdb):
        self.logdb = logdb

    def docs(self, request):
        """Fetch LogDB messages for given request"""
        if  request == 'all':
            docs = self.logdb.get_all_requests()
        else:
            docs = self.logdb.get(request)
        return docs

    def orderby(self, docs, order):
        """Order LogDB messages by given type"""
        odict = {}
        for item in docs:
            odict[item['ts']] = item
        keys = sorted(odict.keys())
        keys.reverse()
        out = []
        for key in keys:
            out.append(odict[key])
        return out

    def to_json(self, request, order='ts'):
        """Represent given messages in JSON data-format for given set of requests"""
        docs = self.orderby(self.docs(request), order)
        return docs

    def to_txt(self, request, order='ts', sep=' '):
        """Represent given messages in ASCII text format for given set of requests"""
        docs = self.orderby(self.docs(request), order)
        keys = list(docs[0])
        out = sep.join(keys) + '\n'
        for doc in docs:
            values = []
            for key in keys:
                values.append('%s' % doc[key])
            out += sep.join(values) + '\n'
        return out

    def to_html(self, request, order='ts'):
        """Represent given messages in ASCII text format for given set of requests"""
        out = '<table id="logdb-report">\n'
        for doc in self.to_txt(request, order, sep='</td><td>').split('\n'):
            if  doc:
                out += '<tr><td>'+doc+'</td></tr>\n'
        out += '</table>'
        return out

    def to_stdout(self, request, order='ts'):
        """Yield to stdout LogDB messages for given request/type"""
        msg = '\nReport for %s' % request
        print(msg, '\n', '-'*len(msg))
        docs = self.orderby(self.docs(request), order)
        times = []
        messages = []
        mtypes = []
        for doc in docs:
            times.append(str(doc['ts']))
            messages.append(doc['msg'])
            mtypes.append(doc['type'])
        tstpad = max([len(t) for t in times])
        msgpad = max([len(m) for m in messages])
        mtppad = max([len(m) for m in mtypes])
        out = []
        for idx in range(len(times)):
            tcol = '%s%s' % (times[idx], ' '*(tstpad-len(times[idx])))
            mcol = '%s%s' % (messages[idx], ' '*(msgpad-len(messages[idx])))
            ecol = '%s%s' % (mtypes[idx], ' '*(mtppad-len(mtypes[idx])))
            print("%s %s %s" % (tcol, mcol, ecol))



#!/usr/bin/env python
"""
_LogArchive_

Diagnostic implementation for a job's LogArchive step
"""

from builtins import range
import os

from WMCore.WMSpec.Steps.Diagnostic import Diagnostic, DiagnosticHandler

class LAExceptionHandler(DiagnosticHandler):
    """
    _LAExceptionHandler_

    Generic handler for the LogArchive step

    I have no idea what this should do

    """
    def __call__(self, errCode, executor, **args):
        """
        _operator()_

        Twiddle thumbs, contemplate navel, toss coin

        """
        msg         = "Error in LogArchive: %s\n" % (errCode)
        description = "Misc. LogArchive error: %s\n" % (errCode)
        if args.get('ex', False):
            msg += str(args.get('ex'))
        if args.get('ExceptionInstance', False):
            msg += str(args.get('ExceptionInstance'))

        jobRepXml = os.path.join(executor.step.builder.workingDir,
                                 getattr(executor.step.output, "jobReport", ""))

        if not os.path.exists(jobRepXml):
            # no report => Error
            msg = "No Job Report Found: %s" % jobRepXml
            executor.report.addError(executor.stepName, 50115, "MissingJobReport", msg)
            return

        # job report XML exists, load the exception information from it
        self.parse(executor, jobRepXml)

        # make sure the report has the error in it
        errSection = getattr(executor.report.report, "errors", None)
        if errSection == None:
            msg = "Job Report contains no error report, but LogArchiveManager exited non-zero: %s" % errCode
            executor.report.addError(executor.stepName,
                                     errCode, description, msg)
            executor.report.addError(executor.stepName, 50116, "MissingErrorReport", msg)
            return

        else:
            #check exit code in report is non zero
            if executor.report.report.status == 0:
                msg = "Job Report contains no error report, but LogArchiveManager exited non-zero: %s" % errCode
                executor.report.addError(executor.stepName, 50116, "MissingErrorReport", msg)
            executor.report.addError(executor.stepName,
                                     errCode, description, msg)
        return

class LogArchive(Diagnostic):

    def __init__(self):
        Diagnostic.__init__(self)

        # Setup a default handler
        catchAll            = LAExceptionHandler()
        self.defaultHandler = catchAll
        for x in range(0, 255):
            if x not in self.handlers:
                self.handlers.__setitem__(x, catchAll)



#!/usr/bin/env python
"""
_ListForSubmitter_

MySQL function to list jobs for submission
"""

from WMCore.Database.DBFormatter import DBFormatter


class ListForSubmitter(DBFormatter):
    """
    _ListForSubmitter_

    List the available jobs in WMBS order by descending subscription priority,
    descending workflow priority, and ascending workflow ID.
    """
    sql = """SELECT wmbs_job.id AS id,
                    wmbs_job.name AS name,
                    wmbs_job.cache_dir AS cache_dir,
                    wmbs_sub_types.name AS task_type,
                    wmbs_sub_types.priority AS task_prio,
                    wmbs_job.retry_count AS retry_count,
                    wmbs_workflow.name AS request_name,
                    wmbs_workflow.id AS task_id,
                    wmbs_workflow.priority AS wf_priority,
                    wmbs_workflow.task AS task_name
               FROM wmbs_job
               INNER JOIN wmbs_jobgroup ON
                 wmbs_job.jobgroup = wmbs_jobgroup.id
               INNER JOIN wmbs_subscription ON
                 wmbs_jobgroup.subscription = wmbs_subscription.id
               INNER JOIN wmbs_sub_types ON
                 wmbs_subscription.subtype = wmbs_sub_types.id
               INNER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
               INNER JOIN wmbs_workflow ON
                 wmbs_subscription.workflow = wmbs_workflow.id
             WHERE wmbs_job_state.name = 'created'
             ORDER BY
               wmbs_sub_types.priority DESC,
               wmbs_workflow.priority DESC,
               wmbs_workflow.id ASC"""

    limit_sql = " limit %d"

    def execute(self, conn=None, transaction=False, limitRows=None):
        if limitRows:
            extraSql = self.limit_sql % limitRows
        else:
            extraSql = ""

        result = self.dbi.processData(self.sql + extraSql, conn=conn,
                                      transaction=transaction)
        return self.formatDict(result)



"""
_BulkAdd_

MySQL implementation of Fileset.BulkAdd
"""

import time

from WMCore.Database.DBFormatter import DBFormatter

class BulkAdd(DBFormatter):
    """
    _BulkAdd_

    Bulk add multiple files to mupltiple filesets.  The file/fileset mappings
    are passed in as a list of dicts where each dict has two keys:
      fileid
      fileset
    """
    sql = """INSERT INTO wmbs_fileset_files (fileid, fileset, insert_time)
               VALUES (:fileid, :fileset, :timestamp)"""

    sqlAvail = """INSERT INTO wmbs_sub_files_available (subscription, fileid)
                    SELECT wmbs_subscription.id AS subscription, :fileid
                           FROM wmbs_subscription
                    WHERE wmbs_subscription.fileset = :fileset"""

    def execute(self, binds, conn = None, transaction = False):
        timestamp = int(time.time())
        newBinds = []
        for bind in binds:
            newBind = {"timestamp": timestamp}
            newBind.update(bind)
            newBinds.append(newBind)

        self.dbi.processData(self.sql, newBinds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.sqlAvail, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_GetFWJRTaskName_

Oracle implementation of Jobs.GetFWJRTaskName
"""

from WMCore.WMBS.MySQL.Jobs.GetFWJRTaskName import GetFWJRTaskName as MySQLGetFWJRTaskName

class GetFWJRTaskName(MySQLGetFWJRTaskName):
    """
    Identical to MySQL version.
    """
    pass



#!/usr/bin/env python
"""
_Builder_

Interface definition for Step Builder implementations

"""
from __future__ import print_function

from builtins import object
import logging

import WMCore.WMSpec.Steps.BuildTools as BuildTools
from WMCore.WMSpec.ConfigSectionTree import nodeName

taskSpaceInit = "#!/usr/bin/env python\n"
taskSpaceInit += "# StepSpace Init module for Step\n"
taskSpaceInit += "# Autogenerated by WMCore.WMSpec.Steps.Builder\n"
taskSpaceInit += "\n__all__ = []\n\n"
taskSpaceInit += "from WMCore.WMRuntime.Bootstrap import establishStepSpace\n\n"
taskSpaceInit += "def _Locator():\n"
taskSpaceInit += "    pass\n\n"
taskSpaceInit += "args = {}\n"


class Builder(object):
    """
    _Builder_

    base interface for any WMStep Builder

    """

    def __init__(self):
        self.taskName = None
        self.stepName = None
        self.stepDir = None
        self.taskSpaceInitMod = None

    def __call__(self, step, taskName, workingDirectory, **args):
        """
        _operator(step)_


        """
        self.stepName = nodeName(step)
        self.taskName = taskName
        self.coreBuild(step, workingDirectory)
        self.build(step, workingDirectory, **args)

    def coreBuild(self, step, workingDirectory):
        """
        _coreBuild_

        Build standard stuff that is common to all jobs

        """
        logging.info("%s.coreBuild invoked", self.__class__.__name__)
        dirs = BuildTools.makeDirectory(step)
        dirs.create(workingDirectory)
        self.stepDir = "%s/%s" % (workingDirectory, self.stepName)
        #
        # Install the basic __init__.py module into the step
        # directory

        self.taskSpaceInitMod = "%s/__init__.py" % self.stepDir

        with open(self.taskSpaceInitMod, 'w') as handle:
            handle.write(taskSpaceInit)
            handle.write("""args["TaskName"] = "%s"\n""" % self.taskName)
            handle.write("""args["StepName"] = "%s"\n""" % self.stepName)
            handle.write("""args["Locator"] = _Locator\n""")
            handle.write("""stepSpace = establishStepSpace(**args)\n""")

    def build(self, step, workingDirectory, **args):
        """
        _build_

        Build the step into the working area provided
        args is for all the things we havent thought of yet

        """
        msg = "WMSpec.Steps.Builder.build method not overridden in "
        msg += "implementation: %s\n" % self.__class__.__name__
        raise NotImplementedError(msg)

    def installWorkingArea(self, step, workingArea, **args):
        """
        _installWorkingArea_

        Install working directory information into the step in a standard
        way.

        """
        step.section_("builder")
        step.builder.workingDir = workingArea
        return



# -*- coding: utf-8 -*-
"""
_SetRunningJobSlotsForSite_

Oracle implementation of ResourceControl.SetRunningJobSlotsForSite

Created on Fri Jun 22 14:16:24 2012

@author: dballest
"""

from WMCore.ResourceControl.MySQL.SetRunningJobSlotsForSite import \
        SetRunningJobSlotsForSite as MySQLSetRunningJobSlotsForSite

class SetRunningJobSlotsForSite(MySQLSetRunningJobSlotsForSite):
    pass



#!/usr/bin/env python
"""
_LoadForErrorHandler_

Oracle implementation of Jobs.LoadForErrorHandler.
"""

from WMCore.WMBS.MySQL.Jobs.LoadForErrorHandler import LoadForErrorHandler as MySQLLoadForErrorHandler


class LoadForErrorHandler(MySQLLoadForErrorHandler):
    """
    The MySQL query should work, but no, it doesn't...
    """
    fileSQL = """SELECT wfd.id, wfd.lfn, wfd.filesize "size", wfd.events, wfd.first_event,
                   wfd.merged, wja.job "jobid", wpnn.pnn
                 FROM wmbs_file_details wfd
                   INNER JOIN wmbs_job_assoc wja ON wja.fileid = wfd.id
                   LEFT OUTER JOIN wmbs_file_location wfl ON wfd.id = wfl.fileid
                   LEFT OUTER JOIN wmbs_pnns wpnn ON wpnn.id = wfl.pnn
                 WHERE wja.job = :jobid"""

    parentSQL = """SELECT parent.lfn AS lfn, wfp.child AS id
                     FROM wmbs_file_parent wfp
                     INNER JOIN wmbs_file_details parent ON parent.id = wfp.parent
                     WHERE wfp.child = :fileid AND parent.merged = '1'"""



#!/usr/bin/env python
"""
_CountOpenBlocks_

MySQL implementation of DBSBuffer.CountOpenBlocks

"""

from __future__ import division
from WMCore.Database.DBFormatter import DBFormatter


class CountOpenBlocks(DBFormatter):
    """
    _CountOpenBlocks_

    Count the number of open blocks. Used to report open blocks for drain status.

    """
    sql = """SELECT COUNT(*)
             FROM dbsbuffer_block
             WHERE status!='Closed'"""

    def execute(self, conn=None, transaction=False):
        result = self.dbi.processData(self.sql, conn=conn,
                                      transaction=transaction)
        result = self.format(result)
        return result[0][0]
